#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include <asm/msr-index.h>

#include "xenner.h"
#include "processor.h"
#include "cpufeature.h"

/* ------------------------------------------------------------------ */

#include "shared-cpuid.c"

void setup_cpuid(struct xenvm *xen)
{
    struct kvm_cpuid_entry list[] = {
	{ .function = 0x00000001 },
	{ .function = 0x80000001 },
    };
    int count = sizeof(list)/sizeof(list[0]);
    int i;

    for (i = 0; i < count; i++) {
	real_cpuid(list+i);
	filter_cpuid(list+i, xen->mode == XENMODE_64);
    }
    kvm_setup_cpuid(xen->kvm, 0, count, list);
}

/* ------------------------------------------------------------------ */

static const char *msr_low[] = {
    [ MSR_IA32_P5_MC_ADDR              ] = "IA32_P5_MC_ADDR",
    [ MSR_IA32_TSC                     ] = "IA32_TSC",
    [ MSR_IA32_SYSENTER_CS             ] = "IA32_SYSENTER_CS",
    [ MSR_IA32_SYSENTER_ESP            ] = "IA32_SYSENTER_ESP",
    [ MSR_IA32_SYSENTER_EIP            ] = "IA32_SYSENTER_EIP",
};

static const char *msr_high[] = {
    [ MSR_EFER             - 0xc0000000 ] = "EFER",
    [ MSR_STAR             - 0xc0000000 ] = "STAR",
    [ MSR_LSTAR            - 0xc0000000 ] = "LSTAR",
    [ MSR_CSTAR            - 0xc0000000 ] = "CSTAR",
    [ MSR_SYSCALL_MASK     - 0xc0000000 ] = "SYSCALL_MASK",
    [ MSR_FS_BASE          - 0xc0000000 ] = "FS_BASE",
    [ MSR_GS_BASE          - 0xc0000000 ] = "GS_BASE",
    [ MSR_KERNEL_GS_BASE   - 0xc0000000 ] = "KERNEL_GS_BASE",
};

static const char *msr_name(uint32_t index)
{
    if (index > 0xc0000000)
	return GETNAME(msr_high, index - 0xc0000000);
    return GETNAME(msr_low, index);
}

void set_msr(struct xenvm *xen, int vcpu, uint32_t index, uint64_t data)
{
    struct kvm_msr_entry msr = {
	.index    = index,
	.reserved = 0,
	.data     = data,
    };

    d1printf("%s: %" PRIx32 " (%s) = %" PRIx64 "\n",
	     __FUNCTION__, index, msr_name(index), data);
    kvm_set_msrs(xen->kvm, vcpu, &msr, 1);
}

uint64_t get_msr(struct xenvm *xen, int vcpu, uint32_t index)
{
    struct kvm_msr_entry msr = {
	.index    = index,
	.reserved = 0,
    };

    kvm_get_msrs(xen->kvm, vcpu, &msr, 1);
    d3printf("%s: %" PRIx32 " (%s) = %llx\n",
	     __FUNCTION__, index, msr_name(index), msr.data);
    return msr.data;
}

void print_msrs(struct xenvm *xen, int vcpu)
{
    struct kvm_msr_list *msrs;
    struct kvm_msr_entry entry;
    int i;

    fprintf(stderr, "--- msrs (nonzero) ---\n");
    msrs = kvm_get_msr_list(xen->kvm);
    for (i = 0; i < msrs->nmsrs; i++) {
	entry.index = msrs->indices[i];
	kvm_get_msrs(xen->kvm, vcpu, &entry, 1);
	if (0 == entry.data)
	    continue;
	fprintf(stderr, "msr %08x = %016llx (%s)\n",
		entry.index, entry.data, msr_name(entry.index));
    }
}
